'use strict';

const _util = require('sequelize/lib/utils');
const Op = require('sequelize/lib/operators');

const fs = require('fs-extra');
const path = require('path');
const types = require('./types');

const RE_DATA = /^data:(.+?);base64,/;

function _unflattenData(target, props, keys, rfx) {
  Object.keys(props).forEach(key => {
    /* istanbul ignore else */
    if (props[key] !== null) {
      const parts = key.split('.');
      const name = parts.pop();

      let obj = target;

      while (parts.length) {
        const prop = parts.shift();

        /* istanbul ignore else */
        if (!obj[prop]) {
          obj[prop] = {};
        }

        obj = obj[prop];
      }

      obj[name] = props[key];
    }
  });

  keys.forEach(key => {
    /* istanbul ignore else */
    if (rfx[key].hasManyItems && !Array.isArray(target[key])) {
      target[key] = target[key] ? [target[key]] : [];
    }
  });

  return target;
}

function _unflattenObj(pk, names, values, references) {
  const grouped = [];
  const keys = {};

  values.forEach(item => {
    const row = _unflattenData({}, item, names, references);

    /* istanbul ignore else */
    if (typeof keys[row[pk]] === 'undefined') {
      keys[row[pk]] = grouped.length;
      grouped.push(row);
      return;
    }

    const prev = grouped[keys[row[pk]]];

    names.forEach(prop => {
      /* istanbul ignore else */
      if (!Array.isArray(prev[prop])) {
        prev[prop] = [];
      }

      prev[prop] = prev[prop].concat(row[prop]);
    });
  });

  return grouped;
}

function _cleanData(values, model) {
  Object.keys(model.rawAttributes).forEach(key => {
    const prop = model.rawAttributes[key];

    if ((prop.primaryKey && !values[key]) || (prop._autoGenerated && !prop.references) || (prop.references && !values[key])) {
      delete values[key];
    }

    if (prop.references) {
      const { Model, fieldName } = prop.type.options;

      if (values[Model.name] && values[Model.name][fieldName]) {
        values[key] = values[Model.name][fieldName];
      }
    }
  });

  Object.keys(model.associations).forEach(key => {
    const assoc = model.associations[key];
    const target = (assoc.through && assoc.through.model) || assoc.target;

    if (Array.isArray(values[key])) {
      values[key] = values[key].map(x => _cleanData(x, target));
    } else if (values[key]) {
      values[key] = _cleanData(values[key], target);
    }
  });

  return values;
}

// FIXME: investigate the root-cause of pk=null...
function _fixPkNull(row, model) {
  /* istanbul ignore else */
  if (row[model.primaryKeyAttribute] === null && row.null > 0) {
    row.dataValues[model.primaryKeyAttribute] = row.null;
    delete row.null;
  }

  return row;
}

function _isData(value) {
  return typeof value === 'string' && RE_DATA.test(value);
}

function _fixRefs(refs, values) {
  // FIXME: consider apply this recursively?
  Object.keys(refs).forEach(ref => {
    /* istanbul ignore else */
    if (Array.isArray(values[ref]) && refs[ref] && refs[ref].through) {
      const through = typeof refs[ref].through === 'object'
        ? refs[ref].through.model
        : refs[ref].through;

      values[ref] = values[ref].map(item => {
        const sub = item[through];

        delete item[through];

        sub[refs[ref].model] = item;

        return sub;
      });
    }
  });
}

function _fileInfo(baseDir, data, ex) {
  /* istanbul ignore else */
  if (!data) {
    throw new Error('File data is missing');
  }

  /* istanbul ignore else */
  if (Array.isArray(data)) {
    // FIXME: additional fields on uploads
    return data.map(x => _fileInfo(baseDir, x, ex));
  }

  delete ex.$upload;
  return {
    ...ex,
    mtime: (data.lastModifiedDate || new Date()).toGMTString(),
    path: path.relative(baseDir, data.path),
    name: data.name,
    size: data.size,
    type: data.type,
  };
}

function _pushUpload(key, _schema, destFile, payload, metadata, onUploadCallback) {
  if (key === null) {
    Object.assign(payload, metadata);
  } else if (_schema.type === 'string') {
    payload[key] = JSON.stringify(metadata);
  } else {
    payload[key] = metadata;
  }

  /* istanbul ignore else */
  if (onUploadCallback) {
    return onUploadCallback({
      payload,
      metadata,
      destFile,
      field: key,
      schema: _schema,
    });
  }
}

function _saveBase64(key, schema, payload, attachments, onUploadCallback) {
  const baseDir = attachments.baseDir || process.cwd();

  /* istanbul ignore else */
  if (_isData(payload[key])) {
    const details = payload[key].match(RE_DATA)[1].split(';');
    const base64Data = payload[key].replace(RE_DATA, '');

    const fileName = details[1] ? `_${details[1].split('name=')[1]}` : `_${details[0].replace(/\W+/g, '.')}`;
    const destFile = path.join(baseDir, attachments.uploadDir || 'uploads', `${Date.now()}${fileName}`);

    // save file on disk before anything else!
    fs.outputFileSync(destFile, base64Data, 'base64');

    const metadata = {
      mtime: new Date(),
      path: path.relative(baseDir, destFile),
      name: path.basename(destFile),
      size: fs.statSync(destFile).size,
      type: details[0],
    };

    return _pushUpload(key, schema, destFile, payload, metadata, onUploadCallback);
  }
}

function _saveUploads(Model, field, payload, attachments, onUploadCallback) {
  const baseDir = attachments.baseDir || process.cwd();
  const schema = Model.options.$schema.properties;
  const tasks = [];

  /* istanbul ignore else */
  try {
    if (payload.$upload) {
      return _fileInfo(baseDir, attachments.files[payload.$upload], payload);
    }
  } catch (e) {
    throw new Error(`Cannot retrieve ${field} '${payload.$upload}' (${e.message})`);
  }

  if (payload[Model.name]) {
    if (payload[Model.name].$upload) {
      const attachment = _fileInfo(baseDir, attachments.files[payload[Model.name].$upload], payload[Model.name]);

      tasks.push(_pushUpload(Model.name, schema, attachment.path, payload, attachment, onUploadCallback));
    } else {
      Object.keys(Model.associations).forEach(assoc => {
        const data = payload[Model.name][assoc];

        if (data && data.$upload && attachments.files[data.$upload]) {
          const attachment = _fileInfo(baseDir, attachments.files[data.$upload], data);

          tasks.push(_pushUpload(assoc, schema[assoc], attachment.path, payload[Model.name], attachment, onUploadCallback));
        }
      });
    }
  } else {
    Object.keys(Model.rawAttributes).forEach(prop => {
      /* istanbul ignore else */
      if (payload[prop]
        && payload[prop].$upload
        && attachments.files[payload[prop].$upload]
      ) {
        const attachment = _fileInfo(baseDir, attachments.files[payload[prop].$upload], payload[prop]);

        tasks.push(_pushUpload(prop, schema[prop], attachment.path, payload, attachment, onUploadCallback));
      }
    });
  }

  return Promise.all(tasks);
}

// FIXME: too much complexity for updating related rows, just create new ones!!!
function _buildTasks(references, inputData, modelName, _options, models) {
  const tasks = {
    before: [],
    after: [],
  };

  /* istanbul ignore else */
  if (_options.attachments) {
    if (inputData.$upload) {
      const files = _saveUploads(models[modelName], '$upload', inputData, _options.attachments);

      delete inputData.$upload;

      tasks.before.push(() => _pushUpload(null, models[modelName].options.$schema, files.path, inputData, files, _options.upload));
    } else {
      tasks.before.push(() => _saveUploads(models[modelName], null, inputData, _options.attachments, _options.upload));
    }
  }

  function doUpload(schema, prop, item, _ref) {
    /* istanbul ignore else */
    if (_options.attachments) {
      if (item.$upload) {
        const files = _saveUploads(_ref, prop, item, _options.attachments);

        delete item.$upload;

        if (schema.type === 'array') {
          inputData[prop] = [];
          (Array.isArray(files) ? files : [files]).forEach((x, key) => {
            tasks.before.push(() => _pushUpload(key, schema.items || {}, x.path, inputData[prop], x, _options.upload));
          });
        } else {
          tasks.before.push(() => _pushUpload(prop, schema, files.path, inputData, files, _options.upload));
        }
      } else {
        tasks.before.push(() => (_isData(item)
          ? _saveBase64(prop, schema, inputData, _options.attachments, _options.upload)
          : _saveUploads(_ref, prop, item, _options.attachments, _options.upload)));
      }
    }
  }

  Object.keys(inputData).forEach(prop => {
    const _schema = models[modelName].options.$schema.properties[prop];
    const _attr = models[modelName].rawAttributes[prop] || {};

    /* istanbul ignore else */
    if (!_schema) {
      throw new Error(`Missing schema for '${prop}' (given ${JSON.stringify(inputData[prop])})`);
    }

    if (_attr.references && typeof inputData[prop] === 'object') {
      const _ref = _attr.type.options.Model;
      const _pk = _ref.primaryKeyAttribute;

      doUpload(_schema, prop, inputData[prop], _ref);

      tasks.before.push(_opts => _ref.create(inputData[prop], { logging: _opts.logging })
        .then(result => {
          inputData[prop] = _fixPkNull(result, _ref)[_pk];
        }));
    } else {
      const assoc = models[modelName].associations && models[modelName].associations[prop];
      const ref = references[prop];

      /* istanbul ignore else */
      if (ref && assoc) {
        const through = ref.through && typeof ref.through.model === 'string' ? ref.through.model : ref.through;
        const fields = ref.references.primaryKeys.map(x => x.prop);

        const items = !Array.isArray(inputData[prop])
          ? [inputData[prop]]
          : inputData[prop];

        const _ref = models[ref.model];
        const _pk = _ref.primaryKeyAttribute;
        const _attrs = through && models[through].rawAttributes;

        items.filter(Boolean).forEach(item => {
          doUpload(_schema, prop, item, _ref);

          /* istanbul ignore else */
          if (_attrs) {
            Object.keys(_attrs).forEach(field => {
              /* istanbul ignore else */
              if (_attrs[field].references && typeof item[field] === 'object') {
                doUpload(_schema.items || _schema, field, item, models[through]);

                tasks.before.push(_opts => _ref.create(item[field], { logging: _opts.logging })
                  .then(result => {
                    item[field] = _fixPkNull(result, _ref)[_pk];
                  }));
              }
            });
          }

          tasks.after.push((fk, opts, isUpdate) => {
            const logging = opts.logging;

            const _value = item[ref.model];
            const _keys = Object.keys(_ref.primaryKeys);
            const _isNew = _value && !_keys.every(x => _value[x]);
            const _whereOpts = fields.reduce((prev, cur) => {
              /* istanbul ignore else */
              if (item[cur]) {
                prev[cur] = item[cur];
              }

              delete item[cur];
              return prev;
            }, {});

            const props = {
              [assoc.foreignKey]: fk,
            };

            const refs = Object.keys(_ref.associations).reduce((memo, cur) => {
              /* istanbul ignore else */
              if (_value && _value[cur]) memo.push(_ref.associations[cur]);
              return memo;
            }, []);

            /* istanbul ignore else */
            if (!isUpdate && _value && _value[_pk]) {
              item[through ? assoc.otherKey : assoc.foreignKey] = _value[_pk];
            }

            /* istanbul ignore else */
            if (item[assoc.otherKey] && !_whereOpts[_ref.primaryKeyAttribute]) {
              _whereOpts[assoc.otherKey] = item[assoc.otherKey];

              delete item[assoc.otherKey];
            }

            /* istanbul ignore else */
            if (_isNew) {
              /* istanbul ignore else */
              if (through) {
                return _ref.create(_value || {}, { logging, include: refs })
                  .then(x => _fixPkNull(x, _ref))
                  .then(x => models[through].create(Object.assign({ [assoc.otherKey]: x[_pk] }, item, props)), { logging })
                  .then(x => _fixPkNull(x, models[through]));
              }
            } else if (through) {
              /* istanbul ignore else */
              if (!_whereOpts[models[through].primaryKeyAttribute]) {
                _whereOpts[assoc.foreignKey] = props[assoc.foreignKey];

                delete props[assoc.foreignKey];
              }

              /* istanbul ignore else */
              if (isUpdate) {
                return _options.noupdate !== true && models[through]
                  .update(Object.assign(item, props), { where: _whereOpts, logging });
              }

              return models[through]
                .create(Object.assign(item, props, _whereOpts), { logging, include: refs })
                .then(x => _fixPkNull(x, models[through]));
            } else if (isUpdate) {
              return _options.noupdate !== true && _ref
                .update(item, { where: Object.assign(props, _whereOpts), logging, include: refs });
            }

            /* istanbul ignore else */
            if (Array.isArray(inputData[prop]) && inputData[prop].length > 1) {
              return Promise.all(inputData[prop]
                .map(x => _ref.create(Object.assign(x, props), { logging, incude: refs }).then(y => _fixPkNull(y, _ref))));
            }

            return _ref.create(Object.assign(
              (Array.isArray(inputData[prop]) && inputData[prop][0]) || inputData[prop],
              typeof item === 'object' ? item : null,
              props), { logging, include: refs })
              .then(x => _fixPkNull(x, _ref));
          });
        });
      } else if (_schema.attachment && _isData(inputData[prop])) {
        return _saveBase64(prop, _schema, inputData, _options.attachments, _options.upload);
      }
    }
  });

  return tasks;
}

function _mixOptions(source, target) {
  /* istanbul ignore else */
  if (source) {
    Object.keys(source).forEach(key => {
      if (!target[key]) {
        target[key] = source[key];
      } else {
        target[key] = _mixOptions(source[key], target[key] || {});
      }
    });
  }

  return target;
}

function _pushKeys(model, obj) {
  Object.keys(model.rawAttributes).forEach(key => {
    const prop = model.rawAttributes[key];

    /* istanbul ignore else */
    if ((prop.primaryKey || prop.references) && obj.attributes.indexOf(prop.fieldName) === -1) {
      obj.attributes.unshift(prop.fieldName);
    }
  });
}

function _packOptions(action, model, obj) {
  obj.attributes = (obj.attributes || []).slice();
  obj.include = (obj.include || []).slice();

  /* istanbul ignore else */
  if (action.indexOf('find') === 0) {
    _pushKeys(model, obj);
  }

  Object.keys(obj).forEach(key => {
    /* istanbul ignore else */
    if (!Array.isArray(obj[key]) && typeof obj[key] === 'object' && obj[key] !== null && model.associations[key]) {
      const target = model.associations[key].target;
      const props = _packOptions(action, target, obj[key] || {});

      /* istanbul ignore else */
      if (obj.where && obj.where[key]) {
        if (model.associations[key].through) {
          props.through = {
            attributes: Object.keys(obj.where[key]),
            model: model.associations[key].through.model,
            where: obj.where[key],
          };

          _pushKeys(model.associations[key].through.model, props.through);
        } else {
          props.where = obj.where[key];
        }

        delete obj.where[key];
      }

      /* istanbul ignore else */
      if (props.order && typeof props.order[0] === 'string') {
        props.order.unshift({
          model: target,
          as: key,
        });

        obj.order = obj.order || [];

        /* istanbul ignore else */
        if (obj.order[0] && typeof obj.order[0][0] !== 'object') {
          obj.order = [obj.order];
        }

        obj.order.push(props.order);

        delete props.order;
      }

      switch (action) {
        case 'findAll':
        case 'findOne':
          props.model = target;
          props.as = key;

          if (model.associations[key].through) {
            const refName = model.associations[key].through.model.name;
            const srcName = model.associations[key].target.name;

            if (props[refName]) {
              props.include = props.include.concat(props[refName][srcName].include);
              props.attributes = props.attributes.concat(props[refName][srcName].attributes);
            }

            delete props[refName];
          }

          obj.include.push(props);
          break;

        case 'destroy':
          obj.include.push(model.associations[key]);
          break;

        // nothing to do
        default: break;
      }

      delete obj[key];
    }
  });

  return obj;
}

function _getOpts(model, props, action, params) {
  const fields = Object.assign({}, model.options.$attributes || {});
  const attrs = fields[action]
    || fields.findAll
    || [];

  const obj = {
    include: [],
    attributes: [],
  };

  /* istanbul ignore else */
  if (fields.where) {
    params.where = Object.assign({}, params.where, fields.where);
  }

  /* istanbul ignore else */
  if (params.search) {
    const lookup = fields.search || fields.findAll;
    const orWhere = [];

    let query = params.search.replace(/[^^$\w]+/g, '%');

    query = `%${query}%`.replace(/^%\^/, '').replace(/\$%$/, '');

    // FIXME: nested lookup is not well-tested!
    lookup.forEach(key => {
      if (!key.includes('.')) {
        orWhere.push({ [key]: { [Op.like]: query } });
      } else {
        const [ref] = key.split('.');

        params.include = params.include || [];
        params.include.push({
          model: model.associations[ref].target,
          as: ref,
        });

        orWhere.push({ [`$${key}$`]: { [Op.like]: query } });
      }
    });

    delete params.search;
    params.where = params.where || {};
    params.where[Op.or] = orWhere;
  }

  attrs.forEach(field => {
    const key = typeof field === 'object'
      ? field.prop
      : field;

    /* istanbul ignore else */
    if (!key || props[key]) {
      return;
    }

    if (key.indexOf('.') === -1) {
      obj.attributes.push(key);
    } else {
      const keys = key.split('.');

      let u = obj;
      let k;

      while (keys.length > 1) {
        k = keys.shift();

        /* istanbul ignore else */
        if (!u[k]) {
          u[k] = {
            include: [],
            attributes: [],
          };
        }

        u = u[k];
      }

      u.attributes.push(keys[0]);
    }
  });

  return _packOptions(action, model, _mixOptions(params, obj));
}

module.exports = (conn, options, modelName) => {
  /* istanbul ignore else */
  if (typeof options === 'string') {
    modelName = options;
    options = null;
  }

  /* istanbul ignore else */
  if (!(conn && conn.$refs && typeof conn.$refs[modelName] === 'object')) {
    throw new Error(`Invalid model definition, given '${modelName}'`);
  }

  const model = conn.models[modelName];

  options = options || {};
  options.keys = options.keys || [];

  const _where = {};
  const _props = {};

  Object.keys(model.rawAttributes)
    .filter(x => model.rawAttributes[x].primaryKey).sort()
    .forEach((pk, i) => {
      /* istanbul ignore else */
      if (typeof options.keys[i] !== 'undefined') {
        _where[pk] = options.keys[i];
      }
    });

  /* istanbul ignore else */
  if (typeof options.where === 'string') {
    const p = options.where.split(';');

    p.forEach(v => {
      const s = v.split(':').filter(Boolean);

      _where[s[0]] = s.length === 2
        ? s[1]
        : s;
    });
  }

  const _schema = types.cleanSchema(model.options.$schema || {});

  function push(ref, nested) {
    const through = typeof ref.params.through === 'object'
      ? ref.params.through.model
      : ref.params.through;

    // attribute references
    _props[ref.params.as] = {
      through,
      model: ref.target,
      plural: _util.pluralize(ref.target),
      singular: _util.singularize(ref.target),
      references: conn.$refs[ref.target].$references,
      hasManyItems: ref.method.indexOf('Many') > 0,
      requiredProps: conn.$refs[ref.target].$schema.required || [],
    };

    /* istanbul ignore else */
    if (nested > 0) {
      _props[ref.params.as].uiSchema = conn.$refs[ref.target].$uiSchema || {};
    }
  }

  function append(id, nested) {
    /* istanbul ignore else */
    if (!(conn.$refs[id].$references && conn.$refs[id].$dependencies)) {
      return;
    }

    const fields = conn.$refs[id].$schema.properties;

    Object.keys(conn.$refs[id].$dependencies).forEach(dep => {
      const ref = conn.$refs[id].$dependencies[dep];

      push(ref, nested);

      /* istanbul ignore else */
      if (!nested) {
        return;
      }

      // append references from virtual-models
      Object.keys(fields).forEach(prop => {
        const _ref = (fields[prop].items || fields[prop]).$ref;

        /* istanbul ignore else */
        if (conn.$refs[_ref]) {
          _props[_ref] = conn.$refs[_ref].$schema;

          append(_ref, conn.$refs[id].$schema.virtual ? 0 : (nested - 1));
        }
      });
    });
  }

  // model references
  _props[modelName] = Object.assign({
    model: modelName,
  }, model.options.name);

  _props[modelName].references = conn.$refs[modelName].$references;

  append(modelName, 1);

  // shared resource
  const instance = {
    options: {
      model: modelName,
      refs: conn.schemas.reduce((prev, cur) => {
        if (!prev[cur.id]) {
          prev[cur.id] = cur;
        }

        return prev;
      }, _props),
      schema: _schema,
      uiSchema: model.options.$uiSchema || {},
      attributes: model.options.$attributes || {},
    },
  };

  function ok(result) {
    /* istanbul ignore else */
    if (typeof result !== 'undefined' && result !== null) {
      if (Array.isArray(result)) {
        result.forEach(x => _fixRefs(_props, x));
      } else {
        _fixRefs(_props, result);
      }
      return result;
    }
  }

  function err(error) {
    if (options.fallthrough) {
      instance.options.failure = error.message;
    } else {
      throw error;
    }
  }

  function build(payload, isUpdate, _options) {
    _options = _options || {};
    _options.logging = options.logging || _options.logging;
    _options.where = Object.assign({}, _options.where, _where);

    let _payload;
    let _tasks;
    let _opts;

    return Promise.resolve()
      .then(() => {
        _payload = _cleanData(payload || options.payload, model);
        _tasks = _buildTasks(_props, _payload, model.name, options, conn.models);
        _opts = _getOpts(model, _props, isUpdate ? 'update' : 'create', _options);
      })
      .then(() => _tasks.before.reduce((prev, cur) => prev.then(() => cur(_opts, isUpdate)), Promise.resolve()))
      .then(() => model[isUpdate ? 'update' : 'create'](_payload, _opts))
      .then(row => {
        let pk;

        /* istanbul ignore else */
        if (!Array.isArray(row)) {
          pk = _fixPkNull(row, model)[model.primaryKeyAttribute];
        }

        /* istanbul ignore else */
        if (!pk && _opts.where) {
          pk = _opts.where[model.primaryKeyAttribute];
        }

        _payload[model.primaryKeyAttribute] = _payload[model.primaryKeyAttribute] || pk;

        /* istanbul ignore else */
        if (!isUpdate) {
          row = [row, _payload];
        }

        return _tasks.after.reduce((prev, cur) => prev.then(() => cur(pk, _opts, isUpdate)), Promise.resolve()).then(() => row);
      })
      .catch(err);
  }

  function fetch(method, _options) {
    _options = _options || {};
    _options.logging = options.logging || _options.logging;
    _options.where = Object.assign({}, _options.where, _where);

    const _pk = model.primaryKeyAttribute;
    const _opts = _getOpts(model, _props, method, _options);
    const _names = _opts.include.map(inc => inc.as);

    // see: https://github.com/sequelize/sequelize/pull/9384
    _opts.raw = true;
    _opts.plain = false;

    return Promise.resolve()
      .then(() => model[method](_opts))
      .then(data => {
        /* istanbul ignore else */
        if (method.indexOf('find') === 0) {
          const result = _unflattenObj(_pk, _names, data, instance.options.refs);

          return method === 'findOne' ? result[0] : result;
        }

        return data;
      })
      .then(ok)
      .catch(err);
  }

  instance.actions = {
    update: (payload, _options) => build(payload, true, _options),
    findAll: _options => fetch('findAll', _options),
    findOne: _options => fetch('findOne', _options),
    destroy: _options => fetch('destroy', _options),
    count: _options => fetch('count', _options),
    create: (payload, _options) => build(payload, false, _options),
  };

  return instance;
};
